ArenaHistory.Player = {}
local Player = ArenaHistory.Player
local PlayerSpec = ArenaHistory.PlayerSpec
local Logger = ArenaHistory.Logger
local GameProto = ArenaHistory.GameProto

Player.__index = Player
Player.logger = Logger.new("Player")

local LSR = LibStub("LibSpecRoster-1.0")
local PB = LibStub("pblua").compile(GameProto.ast)
local DRData = LibStub:GetLibrary("DRData-1.0")

local MAX_DEBUFFS = 40;

function Player.new()
  self = setmetatable({}, Player)
  self.id = nil
  self.guid = nil
  self.name = nil
  self.server = nil
  self.race = nil
  self.faction = nil
  self.specId = nil
  self.damageDone = nil
  self.healingDone = nil
  self.ratingChange = nil

  -- cc
  self.time_in_cc = 0
  self.cc_start_timestamp = nil

  self.died_timestamp = nil

  return self
end

function Player:merge(player)
  self.id = self.id or player.id
  self.guid = self.guid or player.guid
  self.name = self.name or player.name
  self.server = self.server or player.server
  self.race = self.race or player.race
  self.faction = self.faction or player.faction
  self.specId = self.specId or player.specId
  self.damageDone = self.damageDone or player.damageDone
  self.healingDone = self.healingDone or player.healingDone
  self.ratingChange = self.ratingChange or player.ratingChange
  if self.time_in_cc == 0 then
    self.time_in_cc = player.time_in_cc
  end
end

function Player:getFullName()
  if self.name == nil or self.server == nil then
    return nil
  else
    return self.name .. "-" .. self.server
  end
end

function Player:hasAllData()
  return self.id ~= nil and self.name ~= nil and self.server ~= nil and
     self.specId ~= nil and self.faction ~= nil
end

function Player.isArenaId(id)
  if id ~= nil then
    local numberStr = strmatch(id, "^arena(%d+)$")
    if numberStr then
      return tonumber(numberStr)
    end
  end
  return nil
end

function Player:getArenaId()
  return Player.isArenaId(self.id)
end

function Player:update()
  if self.id == nil or self:hasAllData() then
    return
  end
  local id = self.id
  self.guid = self.guid or UnitGUID(id)
  local name, server = Player.getNameAndServer(
      GetUnitName(id, true  --[[ show server name --]]))
  self.name = self.name or name
  self.server = self.server or server
  self.race = self.race or select(2, UnitRace(id))
  self.faction = self.faction or select(1, UnitFactionGroup(id))
  if self.specId == nil then
    local arenaId = self:getArenaId()
    if arenaId then
      self.specId = GetArenaOpponentSpec(arenaId)
    else
      self.specId = select(2, LSR:getSpecialization(UnitGUID(id)))
    end
  end
end

function Player.getNameAndServer(nameAndServer)
  if nameAndServer == nil or nameAndServer == UNKNOWNOBJECT then
    return nil, nil
  end
  local name, server = nameAndServer:match("^(%s*[^- ]*)%s*%-%s*(.*)%s*$")
  if name == nil or server == nil then
    name = nameAndServer
    server = GetRealmName()
  end
  return name, server
end

function Player:addBattlefieldScore(index)
  local name, killingBlows, honorableKills, deaths, honorGained, faction, race,
      class, classToken, damageDone, healingDone, bgRating, ratingChange,
      preMatchMMR, mmrChange, talentSpec = GetBattlefieldScore(index)
  Player.logger:debug("Adding scores for player %s", name)
  self.name = self.name or select(1, Player.getNameAndServer(name))
  self.server = self.server or select(2, Player.getNameAndServer(name))
  self.damageDone = damageDone
  self.healingDone = healingDone
  self.ratingChange = ratingChange
  self.specId = self.specId or PlayerSpec:getSpecIdFromLocalizedName(
      classToken, talentSpec)
  self.logger:info("Parsed bg score #%d: %s", index, self:debugString())
end

function Player:debugString()
  return string.format("id: %s; name: %s; server: %s; spec: %d;",
      self.id or "null", self.name or "null", self.server or "null", self.specId or -1)
end

function Player:isInCC()
  for i = 1, MAX_DEBUFFS do
    local _, _, _, _, _, _, _, _, _, _, spellID = UnitDebuff(self.id, i);
    if (spellID ~= nil and DRData:GetSpellCategory(spellID)) then
      return true
    end
  end
  return false
end

function Player:isFakingDeath()
  local fakeDeathSpells = {5384}
  for _, spellId in pairs(fakeDeathSpells) do
    local name = select(1, GetSpellInfo(spellId))
    if name ~= nil and select(1, UnitAura(self.id, name)) ~= nil then
      self.logger:info("Player %s (%s) is faking death!",
          self.id or "?", self.name or "?")
      return true
    end
  end
  return false
end

function Player:updateAuras()
  local cced = self:isInCC()
  if not cced then
    self:clearCC()
  elseif self.cc_start_timestamp == nil then
    self.cc_start_timestamp = GetTime()
  end
end

function Player:died()
  if not self:isFakingDeath() then
    self:clearCC()
    self.died_timestamp = time()
  end
end

function Player:clearCC()
  if self.cc_start_timestamp ~= nil then
    local time_spent = GetTime() - self.cc_start_timestamp
    self.logger:debug("Player %s spent %fs in CC", self.id, time_spent)
    self.time_in_cc = self.time_in_cc + time_spent 
    self.cc_start_timestamp = nil
  end
end

function Player:dump(isProto)
  local dump
  if isProto then
    dump = PB.Player()
  else
    dump = {}
  end
  dump.id = self.id
  dump.name = self.name
  dump.server = self.server
  dump.race = self.race
  dump.faction = self.faction
  dump.spec_id = self.specId
  dump.damage_done = self.damageDone
  dump.healing_done = self.healingDone
  dump.rating_change = self.ratingChange
  dump.time_in_cc = math.floor(self.time_in_cc + 0.5)

  return dump
end

function Player.copy(dump, isProto)
  local copy
  if isProto then
    copy = PB.Player()
  else
    copy = {}
  end
  copy.id = dump.id
  copy.name = dump.name
  copy.server = dump.server
  copy.race = dump.race
  copy.faction = dump.faction
  copy.spec_id = dump.spec_id
  copy.damage_done = dump.damage_done
  copy.healing_done = dump.healing_done
  copy.rating_change = dump.rating_change
  copy.time_in_cc = dump.time_in_cc
  return copy
end
